# -*- coding: utf-8 -*-

""" Schema loading and validation.
"""

from __future__ import unicode_literals

import base64
import json
import pkgutil
from typing import Any, Dict, Hashable, Optional, Sequence, Text, TextIO
from copy import deepcopy

import jsonschema
from future.builtins import map

from clkhash.backports import raise_from
from clkhash.field_formats import FieldSpec, spec_from_json_dict, InvalidSchemaError
from clkhash.key_derivation import DEFAULT_KEY_SIZE as DEFAULT_KDF_KEY_SIZE

MASTER_SCHEMA_FILE_NAMES = {1: 'v1.json',
                            2: 'v2.json',
                            3: 'v3.json'}  # type: Dict[Hashable, Text]


class SchemaError(Exception):
    """ The user-defined schema is invalid.
    """

    def __init__(self,
                 msg,                               # type: str
                 errors=None                        # type: Optional[Sequence[InvalidSchemaError]]
                 ):
        # type: (...) -> None
        self.msg = msg
        self.errors = [] if errors is None else errors
        super(SchemaError, self).__init__(msg)

    def __str__(self):
        detail = ""
        for i, e in enumerate(self.errors, start=1):
            detail += "Error {} in feature at index {} - {}\n".format(i, e.field_spec_index, str(e))
            detail += "Invalid spec:\n{}\n---\n".format(e.json_field_spec)

        return self.msg + '\n\n' + detail


class MasterSchemaError(Exception):
    """ Master schema missing? Corrupted? Otherwise surprising? This is
        the exception for you!
    """


class Schema:
    """Linkage Schema which describes how to encode plaintext identifiers.

    :ivar fields: the features or field definitions
    :ivar int l: The length of the resulting encoding in bits. This is the length after XOR folding.
    :ivar int xor_folds: The number of XOR folds to perform on the hash.
    :ivar str kdf_type: The key derivation function to use. Currently,
        the only permitted value is 'HKDF'.
    :ivar str kdf_hash: The hash function to use in key derivation. The
        options are 'SHA256' and 'SHA512'.
    :ivar bytes kdf_info: The info for key derivation. See documentation
        of :func:`key_derivation.hkdf` for details.
    :ivar bytes kdf_salt: The salt for key derivation. See documentation
        of :func:`key_derivation.hkdf` for details.
    :ivar int kdf_key_size: The size of the derived keys in bytes.
    """

    def __init__(self,
                 fields,                            # type: Sequence[FieldSpec]
                 l,                                 # type: int
                 xor_folds=0,                       # type: int
                 kdf_type='HKDF',                   # type: str
                 kdf_hash='SHA256',                 # type: str
                 kdf_info=None,                     # type: Optional[bytes]
                 kdf_salt=None,                     # type: Optional[bytes]
                 kdf_key_size=DEFAULT_KDF_KEY_SIZE  # type: int
                 ):
        # type: (...) -> None
        self.fields = fields
        self.l = l
        self.xor_folds = xor_folds

        self.kdf_type = kdf_type
        self.kdf_type = kdf_type
        self.kdf_hash = kdf_hash
        self.kdf_info = kdf_info
        self.kdf_salt = kdf_salt
        self.kdf_key_size = kdf_key_size

    def __repr__(self):
        return "<Schema (v3): {} fields>".format(len(self.fields))


def _convert_v1_to_v2(schema_dict):
    # type: (Dict[str, Any]) -> Dict[str, Any]
    """
    Convert v1 schema dict to v2 schema dict.
    :param schema_dict: v1 schema dict
    :return: v2 schema dict
    """
    schema_dict = deepcopy(schema_dict)
    version = schema_dict['version']
    if version != 1:
        raise ValueError('Version {} not 1'.format(version))

    clk_config = schema_dict['clkConfig']
    k = clk_config.pop('k')
    clk_hash = clk_config['hash']

    def convert_feature(f):
        if f.get('ignored', False):
            return f

        hashing = f['hashing']
        weight = hashing.get('weight', 1.0)

        if weight == 0:
            return {
                'identifier': f['identifier'],
                'ignored': True
            }

        x = deepcopy(f)
        hashing = x['hashing']
        if 'weight' in hashing:
            del hashing['weight']

        hashing['strategy'] = {}
        hashing['strategy']['k'] = int(round(weight * k))
        hashing['hash'] = clk_hash
        return x

    result = {
        'version': 2,
        'clkConfig': {
            'l': clk_config['l'],
            'xor_folds': clk_config.get('xor_folds', 0),
            'kdf': clk_config['kdf']
        },
        'features': list(map(convert_feature, schema_dict['features']))
    }
    return result


def _convert_v2_to_v3(schema_dict):
    # type: (Dict[str, Any]) -> Dict[str, Any]
    """
    Convert v2 schema dict to v3 schema dict.
    :param schema_dict: v2 schema dict
    :return: v3 schema dict
    """
    schema_dict = deepcopy(schema_dict)
    version = schema_dict['version']
    if version != 2:
        raise ValueError('Version {} not 2'.format(version))
    schema_dict['version'] = 3
    for feature in schema_dict['features']:
        if feature.get('ignored', False):
            continue
        strategy = feature['hashing']['strategy']
        if 'k' in strategy:
            strategy['bitsPerToken'] = strategy.pop('k')
        elif 'numBits' in strategy:
            strategy['bitsPerFeature'] = strategy.pop('numBits')
        ngrams = feature['hashing']['ngram']
        feature['hashing']['comparison'] = {'type': 'ngram', 'n': feature['hashing'].pop('ngram'), 'positional': feature['hashing'].pop('positional', False)}
    return schema_dict


def convert_to_latest_version(schema_dict, validate_result=False):
    # type: (Dict[str, Any], Optional[bool]) -> Dict[str, Any]
    """ Convert the given schema to latest schema version.

     :param schema_dict: A dictionary describing a linkage schema. This dictionary must have a `'version'` key
            containing a master schema version. The rest of the schema dict must conform to the corresponding
            master schema.
     :param validate_result: validate converted schema against schema specification
     :return: schema dict of the latest version
     raises SchemaError if schema version is not supported"""
    version = schema_dict.get('version', "'not specified'")
    if version not in MASTER_SCHEMA_FILE_NAMES.keys():
        msg = ('Schema version {} is not supported. '
               'Consider updating clkhash.').format(version)
        raise SchemaError(msg)
    if schema_dict['version'] == 1:
        schema_dict = _convert_v1_to_v2(schema_dict)
    if schema_dict['version'] == 2:
        schema_dict = _convert_v2_to_v3(schema_dict)
    if validate_result:
        validate_schema_dict(schema_dict)
    return schema_dict


def from_json_dict(dct, validate=True):
    # type: (Dict[str, Any], bool) -> Schema
    """ Create a Schema of the most recent version according to dct

    if the provided schema dict is of an older version, then it will be automatically converted to the latest.

    :param dct: This dictionary must have a `'features'`
            key specifying the columns of the dataset. It must have
            a `'version'` key containing the master schema version
            that this schema conforms to. It must have a `'hash'`
            key with all the globals.
    :param validate: (default True) Raise an exception if the
            schema does not conform to the master schema.
    :raises SchemaError: An exception containing details about why
            the schema is not valid.
    :return: the Schema
    """
    if validate:
        # This raises iff the schema is invalid.
        validate_schema_dict(dct)
    dct = convert_to_latest_version(dct)
    if validate:
        validate_schema_dict(dct)
    clk_config = dct['clkConfig']
    l = clk_config['l']
    xor_folds = clk_config.get('xor_folds', 0)

    kdf = clk_config['kdf']
    kdf_type = kdf['type']
    kdf_hash = kdf.get('hash', 'SHA256')
    kdf_info_string = kdf.get('info')
    kdf_info = (base64.b64decode(kdf_info_string)
                if kdf_info_string is not None
                else None)
    kdf_salt_string = kdf.get('salt')
    kdf_salt = (base64.b64decode(kdf_salt_string)
                if kdf_salt_string is not None
                else None)
    kdf_key_size = kdf.get('keySize', DEFAULT_KDF_KEY_SIZE)

    # Try to parse each feature config and store any errors encountered
    # for reporting.
    feature_errors = []
    feature_configs = []

    for i, feature_config in enumerate(dct['features']):
        try:
            feature_configs.append(spec_from_json_dict(feature_config))
        except InvalidSchemaError as e:
            e.field_spec_index = i
            e.json_field_spec = feature_config
            feature_errors.append(e)

    if len(feature_errors):
        raise SchemaError("Schema was invalid", feature_errors)

    return Schema(feature_configs, l, xor_folds,
                  kdf_type, kdf_hash, kdf_info, kdf_salt, kdf_key_size)


def from_json_file(schema_file, validate=True):
    # type: (TextIO, bool) -> Schema
    """ Load a Schema object from a json file.

        :param schema_file: A JSON file containing the schema.
        :param validate: (default True) Raise an exception if the
            schema does not conform to the master schema.
        :raises SchemaError: When the schema is invalid.
        :return: the Schema

    """
    try:
        schema_dict = json.load(schema_file)
    except ValueError as e:  # In Python 3 we can be more specific
        # with json.decoder.JSONDecodeError,
        # but that doesn't exist in Python 2.
        msg = 'The schema is not a valid JSON file.'
        raise_from(SchemaError(msg), e)

    return from_json_dict(schema_dict, validate=validate)


def _get_master_schema(version):
    # type: (Hashable) -> dict
    """ Loads the master schema of given version

        :param version: The version of the master schema whose path we
            wish to retrieve.
        :raises SchemaError: When the schema version is unknown. This
            usually means that either (a) clkhash is out of date, or (b)
            the schema version listed is incorrect.
        :raises MasterSchemaError: When the master schema is invalid.
        :return: Dict object of the (json) master schema.
    """
    try:
        file_name = MASTER_SCHEMA_FILE_NAMES[version]
    except (TypeError, KeyError) as e:
        msg = ('Schema version {} is not supported. '
               'Consider updating clkhash.').format(version)
        raise_from(SchemaError(msg), e)

    try:
        schema_bytes = pkgutil.get_data('clkhash', 'schemas/{}'.format(file_name))
        if schema_bytes is None:
            msg = ('The master schema could not be loaded. The schema cannot be '
                   'validated. Please file a bug report.')
            raise MasterSchemaError(msg)
    except IOError as e:  # In Python 3 we can be more specific with
        # FileNotFoundError, but that doesn't exist in
        # Python 2.
        msg = ('The master schema could not be found. The schema cannot be '
               'validated. Please file a bug report.')
        raise_from(MasterSchemaError(msg), e)

    try:
        master_schema = json.loads(schema_bytes.decode('utf-8'))
        return master_schema
    except ValueError as e:
        # In Python 3 we can be more specific with
        # json.decoder.JSONDecodeError, but that
        # doesn't exist in Python 2.
        msg = ('The master schema is not a valid JSON file. The schema cannot '
               'be validated. Please file a bug report.')
        raise_from(MasterSchemaError(msg), e)


def validate_schema_dict(schema):
    # type: (Dict[str, Any]) -> None
    """ Validate the schema.

        This raises iff either the schema or the master schema are
        invalid. If it's successful, it returns nothing.

        :param schema: The schema to validate, as parsed by `json`.
        :raises SchemaError: When the schema is invalid.
        :raises MasterSchemaError: When the master schema is invalid.
    """
    if not isinstance(schema, dict):
        msg = ('The top level of the schema file is a {}, whereas a dict is '
               'expected.'.format(type(schema).__name__))
        raise SchemaError(msg)

    if 'version' in schema:
        version = schema['version']
    else:
        raise SchemaError('A format version is expected in the schema.')

    master_schema = _get_master_schema(version)

    try:
        jsonschema.validate(schema, master_schema)
    except jsonschema.exceptions.ValidationError as e:
        raise_from(SchemaError('The schema is not valid.\n\n' + str(e)), e)
    except jsonschema.exceptions.SchemaError as e:
        msg = ('The master schema is not valid. The schema cannot be '
               'validated. Please file a bug report.')
        raise_from(MasterSchemaError(msg), e)



import csv
import re
import sys
import time
from datetime import datetime

from typing import AnyStr, Callable, cast, Pattern, Sequence, Text, Union, Iterable, Any

from future.utils import raise_from as _raise_from
from mypy_extensions import Arg, DefaultNamedArg, NoReturn

try:
    int_from_bytes = int.from_bytes
except AttributeError:
    import codecs


    def _int_from_bytes(bytes, byteorder, signed=False):
        # type: (Sequence[int], str, bool) -> int
        """ Emulate Python 3's `int.from_bytes`.

            Kudos: https://stackoverflow.com/a/30403242 (with
            modifications)

            :param bytes: The bytes to turn into an `int`.
            :param byteorder: Either `'big'` or `'little'`.
        """
        if signed:
            raise NotImplementedError(
                "Signed integers are not currently supported in this "
                "backport.")

        if byteorder == 'big':
            pass
        elif byteorder == 'little':
            bytes = bytes[::-1]
        else:
            raise ValueError("byteorder must be either 'little' or 'big'")

        hex_str = codecs.encode(bytes, 'hex')  # type: ignore
        return int(hex_str, 16)

    # Make this cast since Python 2 doesn't have syntax for default
    # named arguments. Hence, must cast so Mypy thinks it matches the
    # original function.
    int_from_bytes = cast(Callable[[Arg(Union[Iterable[int], Any], 'bytes'),
                                    Arg(str, 'byteorder'),
                                    DefaultNamedArg(bool, 'signed')],
                                   int],
                          _int_from_bytes)


def re_compile_full(pattern, flags=0):
    # type: (AnyStr, int) -> Pattern
    """ Create compiled regular expression such that it matches the
        entire string. Calling re.match on the output of this function
        is equivalent to calling re.fullmatch on its input.

        This is needed to support Python 2. (On Python 3, we would just
        call re.fullmatch.)
        Kudos: https://stackoverflow.com/a/30212799

        :param pattern: The pattern to compile.
        :param flags: Regular expression flags. Refer to Python
            documentation.

        :returns: A compiled regular expression.
    """
    # Don't worry, this short-circuits.
    assert type(pattern) is str or type(pattern) is unicode  # type: ignore

    return re.compile(r'(?:{!s})\Z'.format(pattern), flags=flags)


def _utf_8_encoder(unicode_csv_data):
    return (line.encode('utf-8') for line in unicode_csv_data)


def _p2_unicode_reader(unicode_csv_data, dialect=csv.excel, **kwargs):
    """ Encode Unicode as UTF-8 and parse as CSV.

        This is needed since Python 2's `csv` doesn't do Unicode.

        Kudos: https://docs.python.org/2/library/csv.html#examples

        :param unicode_csv_data: The Unicode stream to parse.
        :param dialect: The CSV dialect to use.
        :param kwargs: Any other parameters to pass to csv.reader.

        :returns: An iterator
    """
    # Encode temporarily as UTF-8:
    utf8_csv_data = _utf_8_encoder(unicode_csv_data)

    # Now we can parse!
    csv_reader = csv.reader(utf8_csv_data, dialect=dialect, **kwargs)

    # Decode UTF-8 back to Unicode, cell by cell:
    return ([unicode(cell, 'utf-8') for cell in row] for row in csv_reader)


unicode_reader = (_p2_unicode_reader  # Python 2 with hacky workarounds.
                  if sys.version_info < (3, 0)
                  else csv.reader)  # Py3 with native Unicode support.

if hasattr(__builtins__, 'TimeoutError'):
    TimeoutError = __builtins__.TimeoutError    # type: type
else:
    TimeoutError = OSError

if sys.version_info > (3, 2):
    strftime = datetime.strftime
else:
    _YEAR_LEN = 4

    # Detect the unsupported '%s' format. But don't match if there's an
    # even number of '%'s before the 's' because those are all escaped.
    _illegal_s = re.compile(r'((^|[^%])(%%)*%s)')


    def _findall(text, substr):
        # Also finds overlaps
        i = 0
        while True:
            j = text.find(substr, i)
            if j == -1:
                return

            yield j
            i = j + 1


    def strftime(dt, fmt):
        # type: (datetime, Text) -> Text
        """ strftime that support years < 1900 in Python < 3.2.

            Kudos: https://github.com/ActiveState/code/blob/master/recipes/Python/306860_proleptic_Gregoridates_strftime_before/recipe-306860.py
        """
        if _illegal_s.search(fmt):
            msg = "this strftime implementation does not handle '%s'"
            raise ValueError()
        if dt.year > 1900:
            return dt.strftime(fmt)

        year = dt.year
        timetuple = dt.timetuple()
        timetuple_without_year = timetuple[1:]

        # Every 28 years the calendar repeats, except through century
        # leap years where it's 6 years.
        # For every non-leap year century, advance by
        # 6 years to get into the 28-year repeat cycle
        delta = 2000 - year
        off = 6 * (delta // 100 + delta // 400)
        year = year + off

        # `year` and `year + (2000 - year) // 28 * 28` have the same
        # layout
        year = year + (2000 - year) // 28 * 28
        # Format with a supported year and look for all occurences of
        # said year.
        year_str = str(year)
        assert len(year_str) == _YEAR_LEN
        s1 = time.strftime(fmt, (year,) + timetuple_without_year)
        sites1 = set(_findall(s1, year_str))

        # Format with another supported year and look again for all
        # occurences of that year.
        year_p28_str = str(year + 28)
        assert len(year_p28_str) == _YEAR_LEN
        s2 = time.strftime(fmt, (year + 28,) + timetuple_without_year)
        sites2 = set(_findall(s2, year_p28_str))

        # Where those coincide is where the year goes according to our
        # format.
        sites = sites1 & sites2

        # We found the year. Now we replace.
        s = s1
        syear = '{:04}'.format(dt.year)
        assert len(syear) == _YEAR_LEN
        for site in sites:
            s = s[:site] + syear + s[site + _YEAR_LEN:]
        return s

# Help MyPy understand that this always throws.
raise_from = cast(Callable[[BaseException, BaseException], NoReturn],
                  _raise_from)
